package com.timestored;

import java.io.BufferedReader;
import java.io.File;
import java.io.IOException;
import java.io.InputStreamReader;
import java.lang.reflect.Method;
import java.net.MalformedURLException;
import java.net.URI;
import java.net.URL;
import java.net.URLClassLoader;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashMap;
import java.util.List;
import java.util.Locale;
import java.util.Map;
import java.util.Random;

import javax.tools.Diagnostic;
import javax.tools.DiagnosticListener;
import javax.tools.JavaCompiler;
import javax.tools.JavaFileObject;
import javax.tools.SimpleJavaFileObject;
import javax.tools.StandardJavaFileManager;
import javax.tools.ToolProvider;

import com.timestored.QjInterpreter.InMemoryJavaFileObject;

/**
 * A poor mans implementation of kdb in java.
 * <br/><br/>
 * Gives a fair idea of how kdb is implemented behind the scenes. 
 * <ul><li>Many functions to handle the different types 
 *  	(in C I would use macros to generate the many variations)</li>
 *  	<li>All the functions here overwrite the array where they can, obviously we could wrap this
 *  		as a K object with a count of how many times its referenced, then only overwrite when count=0.
 *  		Which I believe is what kdb does</li>
 * </ul>
 */
public class Jkdb {
	
	private static final Map<String,Object> data = new HashMap<String, Object>(); 
	private static final Random p = new Random();

	/** Store a value under a given variable name */
	public static int[] set(String key, int[] value) {data.put(key,copy(value)); return value;}
	public static double[] set(String key, double[] value) {data.put(key,copy(value)); return value;}
	public static int set(String key, int value) {data.put(key,value); return value;}
	public static double set(String key, double value) {data.put(key,value); return value;}
	
	/** Retrieve a value for a given variable name */
	public static Object get(String key) {return data.get(key);}
	public static int geti(String key) {return (int)(Integer)data.get(key);}
	public static double getd(String key) {return (double)(Double)data.get(key);}
	public static int[] getia(String key) {return copy((int[])data.get(key));}
	public static double[] getda(String key) {return copy((double[])data.get(key));}

	/** Check Length */
	private static void cl(double[]x, double[]y) {
		if(x.length!=y.length) {throw new IllegalArgumentException("length");}; }
	private static void cl(boolean[]x, boolean[]y) {
		if(x.length!=y.length) {throw new LengthKException();}; }
	private static void cl(int[]x, int[]y) {
		if(x.length!=y.length) {throw new IllegalArgumentException("length");}; }
	/** Check Length */
	private static void cn(Object o) {
		if(o==null) {throw new TypeKException();}; }
	/** Convert to Double */
	private static double[] cd(int[]x) {
		double[] r = new double[x.length]; for(int i=0; i<x.length; i++) { r[i] = x[i]; } return r; }

	private static int[] cI(List<Integer> l) {
		int[] r = new int[l.size()];
		for (int i = 0; i < r.length; i++) { r[i] = l.get(i); }
		return r;
	}

	/** multiply y by every x and return new result **/
	public static double mul(double x, double y) {  return x*y; };
	public static double[] mul(double[] x, double y) { 
		for(int i=0; i<x.length; i++) { x[i]=x[i]*y; } return x; }
	public static double[] mul(double y, double[] x) { return mul(x,y);} 
	public static double[] mul(double[] x, double[] y) {
		cl(x,y); for(int i=0; i<x.length; i++) { x[i]=x[i]*y[i]; } return x; }
	public static double[] mul(double y, int[] x) { return mul(x,y);} 
	public static double[] mul(int[] y, double x) { return mul(cd(y),x);} 
	public static double[] mul(int[] y, int[] x) { return mul(cd(y),cd(x));} 
	
	/** Add y to every x and return new result **/
	public static double add(double x, double y) {  return x+y; };
	static double[] add(double[] x, double y) {
		for(int i=0; i<x.length; i++) { x[i]=x[i]+y; } return x; }
	public static double[] add(double y, double[] x) { return add(x,y);} 
	public static double[] add(double[] x, double[] y) {
		cl(x,y); for(int i=0; i<x.length; i++) { x[i]=x[i]+y[i]; } return x;
	}
	public static double[] add(double y, int[] x) { return add(x,y);} 
	public static double[] add(int[] y, double x) { return add(cd(y),x);} 
	public static double[] add(int[] y, int[] x) { return add(cd(y),cd(x));} 


	/** choose - same as ? in kdb **/
	public static double[] choose(int x, double y) { 
		double[] res = new double[x];
		for(int i=0; i<x; i++) { res[i]=p.nextDouble()*y; } return res; }
	public static int[] choose(int x, int y) { 
		int[] res = new int[x];
		for(int i=0; i<x; i++) { res[i]=p.nextInt(y); } return res; }
	
	/** same as KDB's til function **/
	public static int[] til(int x) {
		if(x<0) { throw new LimitKException(); };
		int[] r = new int[x]; for(int i=0; i<x; i++) { r[i] = i; } return r; }
	
	/** perform cosine on every elements of array and return new array with result. **/
	public static double[] cos(double[] x) {
		for(int i=0; i<x.length; i++) { x[i] = Math.cos(x[i]); } return x; }
	public static double[] cos(int[] x) { return cos(cd(x)); }
	public static double cos(double x) { return Math.cos(x); }
	public static double cos(int x) { return Math.cos(x); }
	/** perform sine on every elements of array and return new array with result. **/
	public static double[] sin(double[] x) {
		for(int i=0; i<x.length; i++) { x[i] = Math.sin(x[i]); } return x; }
	public static double[] sin(int[] x) { return sin(cd(x)); }
	public static double sin(double x) { return Math.sin(x); }
	public static double sin(int x) { return Math.sin(x); }
	/** perform tan on every elements of array and return new array with result. **/
	public static double[] tan(double[] x) {
		for(int i=0; i<x.length; i++) { x[i] = Math.tan(x[i]); } return x; }
	public static double[] tan(int[] x) { return tan(cd(x)); }
	public static double tan(double x) { return Math.tan(x); }
	public static double tan(int x) { return Math.tan(x); }
	/** perform arccosine on every elements of array and return new array with result. **/
	public static double[] acos(double[] x) {
		for(int i=0; i<x.length; i++) { x[i] = Math.acos(x[i]); } return x; }
	public static double[] acos(int[] x) { return acos(cd(x)); }
	/** perform arcsine on every elements of array and return new array with result. **/
	public static double[] asin(double[] x) {
		for(int i=0; i<x.length; i++) { x[i] = Math.asin(x[i]); } return x; }
	public static double[] asin(int[] x) { return asin(cd(x)); }
	/** perform arctan on every elements of array and return new array with result. **/
	public static double[] atan(double[] x) {
		for(int i=0; i<x.length; i++) { x[i] = Math.atan(x[i]); } return x; }
	public static double[] atan(int[] x) { return atan(cd(x)); }
	/** Return the square root of all numbers in the list */
	public static double[] sqrt(double[] x) {
		for(int i=0; i<x.length; i++) { x[i] = Math.sqrt(x[i]); } return x; }
	public static double[] sqrt(int[] x) { return sqrt(cd(x)); }
	public static double sqrt(double x) { return Math.sqrt(x); }
	/** perform ascending sort **/
	public static double[] asc(double[] x) { Arrays.sort(x); return x; };
	public static int[] asc(int[] x) { Arrays.sort(x); return x;};
	public static double asc(double x) { return x; }
	public static double asc(int x) { return x; }
	/** @return maximum value **/
	public static double max(double[] x) { double r = Double.NEGATIVE_INFINITY;
		for(int i=0; i<x.length; i++) { if(x[i]>r){ r=x[i]; }; } return r; }
	public static int max(int[] x) {	int r = Integer.MIN_VALUE;
	for(int i=0; i<x.length; i++) { if(x[i]>r){ r=x[i]; }; } return r; }
	public static double max(double x) { return x; }
	public static int max(int x) { return x; }
	/** @return minimum value **/
	public static double min(double[] x) { double r = Double.POSITIVE_INFINITY;
		for(int i=0; i<x.length; i++) { if(x[i]<r){ r=x[i]; }; } return r; }
	public static int min(int[] x) {	int r = Integer.MAX_VALUE;
	for(int i=0; i<x.length; i++) { if(x[i]<r){ r=x[i]; }; } return r; }
	public static double min(double x) { return x; }
	public static double min(int x) { return x; }
	/** perform ascending sort **/
	public static double[] reverse(double[] x) { 
		for(int i = 0; i < x.length / 2; i++) {
			double t = x[i]; x[i] = x[x.length - i - 1];  x[x.length - i - 1] = t; }; 
		return x; };
		public static int[] reverse(int[] x) { 
			for(int i = 0; i < x.length / 2; i++) {
				int t = x[i]; x[i] = x[x.length - i - 1];  x[x.length - i - 1] = t; }; 
			return x; };
	public static double desc(double x) { return x; }
	public static double desc(int x) { return x; }
	
	/** @return a copy of the array. **/
	public static double[] copy(double[] x) {
		double[] r = new double[x.length];
		for(int i=0; i<x.length; i++) { r[i] = x[i]; } return r; }
	public static int[] copy(int[] x) {
		int[] r = new int[x.length];
		for(int i=0; i<x.length; i++) { r[i] = x[i]; } return r; }

	/** perform modulus on every elements of array and return new array with result. **/
	public static double[] mod(double[] x, double y) {
		boolean b = y==0;
		for(int i=0; i<x.length; i++) { x[i]= b ? Double.NaN : x[i] % y; } return x; }
	public static int[] mod(int[] x, int y) {
		if(y==0) { throw new IllegalArgumentException(); };
		for(int i=0; i<x.length; i++) { x[i]= x[i] % y; } return x; }
	

	/**  @return the sum of the items in its argument.  */
	public static double sum(double[] x) {
		double r=0; for(double d : x) { if(!Double.isNaN(d)) { r+=d; } } return r; }
	public static double sum(double x) {return x; }
	public static int sum(int[] x) {
		int r=0; for(int d : x) { r+=d; } return r; }
	public static int sum(int x) {return x; }
	/**  @return the product of the items in its argument.  */
	public static double prd(double[] x) {
		double r=1; for(double d : x) { r*=d; } return r; }
	public static double prd(double x) {return x; }
	public static int prd(int[] x) {
		int r=1; for(int d : x) { r*=d; } return r; }
	public static int prd(int x) {return x; }

	/**  @return the cumulative sums of the items in its argument.  */
	public static double[] sums(double[] x) {
		for(int i=1; i<x.length; i++) { x[i] = x[i]+x[i-1]; } return x; }
	public static double sums(double x) {return x; }
	/**  @return the cumulative products of the items in its argument.  */
	public static double[] prds(double[] x) {
		for(int i=1; i<x.length; i++) { x[i] = x[i]*x[i-1]; } return x; }
	public static double prds(double x) {return x; }

	/** @return indices of the true elements */
	public static int[] where(boolean[] x) {
		cn(x);
		List<Integer> r = new ArrayList<Integer>(x.length);
		for(int i=0; i<x.length; i++) {	if(x[i]) { r.add(i);}}
		return cI(r); }
	

	/** @return a boolean[] with true elements where nums[index] = val */
	public static boolean[] equal(int x, int[] y) {
		boolean[] r = new boolean[y.length]; 
		for(int i=0; i<y.length; i++) { r[i]= y[i]==x; } 
		return r; }
	
	/**  @return maximum of its arguments. applies to all data types except symbol.  */
	public static boolean[] or(boolean[] x, boolean[] y) {
		cn(x); cn(y); cl(x,y);
		for(int i=0; i<y.length; i++) { x[i]|=y[i];}; return x; }
	/**  @return minimum of its arguments. applies to all data types except symbol.  */
	public static boolean[] and(boolean[] x, boolean[] y) {
		cn(x); cn(y); cl(x,y);
		for(int i=0; i<y.length; i++) { x[i]&=y[i];}; return x; }
	

	/** @return items at the selected indices. */
	public static double[] index(double[] vals, int[] indices) {
		double[] r = new double[indices.length]; 
		for(int i=0; i<r.length; i++) { r[i] = vals[indices[i]];} return r; }
	public static int[] index(int[] vals, int[] indices) {
		int[] r = new int[indices.length]; 
		for(int i=0; i<r.length; i++) { r[i] = vals[indices[i]];} return r; }
	public static boolean[] index(boolean[] vals, int[] indices) {
		boolean[] r = new boolean[indices.length]; 
		for(int i=0; i<r.length; i++) { r[i] = vals[indices[i]];} return r; }


	/**
	 * Returns the largest (closest to positive infinity) double value that 
	 * is not greater than the argument and is equal to a mathematical integer.
	 */
	public static int[] floor(double[] x) {
		int[] r = new int[x.length]; 
		for(int i=0; i<r.length; i++) { r[i] = (int)Math.floor(x[i]);} return r; }
	/** ceiling */
	public static int[] ceiling(double[] x) {
		int[] r = new int[x.length]; 
		for(int i=0; i<r.length; i++) { r[i] = (int)Math.ceil(x[i]);} return r; }
	
	/** abs */
	public static double[] abs(double[] x) {
		for(int i=0; i<x.length; i++) { x[i] = Math.abs(x[i]);} return x; }
	public static int[] abs(int[] x) {
		for(int i=0; i<x.length; i++) { x[i] = Math.abs(x[i]);} return x; }


	
	/** @return last entry in a list or null if the list is empty. */
	public static int last(int[] x) { return x[x.length-1];}
	public static double last(double[] x) { return x[x.length-1];}
	
	
	/** @return the first entry in a list or null if the list is empty. */
	public static int first(int[] x) { return x[0];}
	public static double first(double[] x) { return x[0];}

	
	public static class TypeKException extends IllegalArgumentException {	}
	public static class LengthKException extends IllegalArgumentException { 
		public LengthKException() {	super("length");	}
	}
	public static class LimitKException extends IllegalArgumentException {}
	

	public static void main(String[] args) throws Exception {
		
	    InputStreamReader isr = new InputStreamReader(System.in);
	    BufferedReader br = new BufferedReader(isr);

		File f = new File(QjInterpreter.classOutputFolder);
		f.mkdirs();
	      
		String query = "";
		do {

			if(!query.trim().startsWith("/")) {
			
				try {
									
					String PRE = "package com.timestored; import static com.timestored.Jkdb.*; import java.util.Arrays;"
					+ "public class Testy {  " +
					"" +
					"	public static void p(Object o) {" +
					"			String s = (o instanceof double[]) ? Arrays.toString((double[])o)" +
					" 				: (o instanceof int[]) ? Arrays.toString((int[])o)" +
					"				: (o instanceof boolean[]) ? Arrays.toString((boolean[])o)" +
					"				: o.toString(); " + 
					"	if(s.length()>80) { s = s.substring(0,80) + \"...\"; };" +
					"   System.out.println(s);" +
					"}" +
					"" +
					"public void " +
							"run() {      p(";
					String POST =  "); } }";
					String classStr = PRE + query.replace(";",",").replace("[","(")
							.replace("]",")").replace("=","equal").replace("?", "choose")
							.replace("+", "add").replace("*", "mul").replace("@", "index") + POST;
					
					if(query.trim().length()>0) {
						JavaFileObject file = new InMemoryJavaFileObject(
								"com.timestored.Testy", classStr);
		
						
						// 2.Compile your files by JavaCompiler
						QjInterpreter.deleteDir(f);
						f.mkdirs();
						QjInterpreter.compile(Arrays.<JavaFileObject> asList(file));
		
						// 3.Load your class by URLClassLoader, then instantiate the
						// instance, and call method by reflection
						QjInterpreter.runIt("com.timestored.Testy", "run");
		//				runIt("math.Calculator", "testAdd");
					}
				} catch (Exception exception) {
					exception.printStackTrace();
				}
			}
			System.out.print("jq)");
			query = br.readLine();
			// 1.Construct an in-memory java source file from your dynamic code

		} while (!query.equalsIgnoreCase("\\\\"));
	}
	
}


class QjInterpreter {

	/** where shall the compiled class be saved to (should exist already) */
	public static String classOutputFolder = "tt";

	public static class MyDiagnosticListener implements
			DiagnosticListener<JavaFileObject> {

		public void report(Diagnostic<? extends JavaFileObject> diagnostic) {

//			System.out.println("Line Number->" + diagnostic.getLineNumber());
//			System.out.println("code->" + diagnostic.getCode());
			System.out.println("Message->"
					+ diagnostic.getMessage(Locale.ENGLISH));
//			System.out.println("Source->" + diagnostic.getSource());
//			System.out.println(" ");
		}
	}

	/**
	 * java File Object represents an in-memory java source file <br>
	 * so there is no need to put the source file on hard disk
	 **/
	public static class InMemoryJavaFileObject extends SimpleJavaFileObject {

		private String contents = null;

		public InMemoryJavaFileObject(String className, String contents)
				throws Exception {
			super(URI.create("string:///" + className.replace('.', '/')
					+ Kind.SOURCE.extension), Kind.SOURCE);
			this.contents = contents;
		}

		public CharSequence getCharContent(boolean ignoreEncodingErrors)
				throws IOException {
			return contents;
		}
	}

	/**
	 * Get a simple Java File Object ,<br>
	 * It is just for demo, content of the source code is dynamic in real use
	 * case
	 */
	private static JavaFileObject getJavaFileObject() {
		StringBuilder contents = new StringBuilder("package math;"
				+ "public class Calculator { " + "  public void testAdd() { "
				+ "    System.out.println(200+300); " + "  } "
				+ "  " + "} ");
		JavaFileObject so = null;
		try {
			so = new InMemoryJavaFileObject("math.Calculator",
					contents.toString());
		} catch (Exception exception) {
			exception.printStackTrace();
		}
		return so;
	}

	/** compile your files by JavaCompiler */
	public static void compile(Iterable<? extends JavaFileObject> files) {
		// get system compiler:
		JavaCompiler compiler = ToolProvider.getSystemJavaCompiler();

		// for compilation diagnostic message processing on compilation
		// WARNING/ERROR
		MyDiagnosticListener c = new MyDiagnosticListener();
		StandardJavaFileManager fileManager = compiler.getStandardFileManager(
				c, Locale.ENGLISH, null);
		// specify classes output folder
		Iterable options = Arrays.asList("-d", classOutputFolder);
		JavaCompiler.CompilationTask task = compiler.getTask(null, fileManager,
				c, options, null, files);
		Boolean result = task.call();
//		if (result == true) {
//			System.out.println("Succeeded");
//		}
	}

	/** run class from the compiled byte code file by URLClassloader 
	 * @param methodName */
	public static void runIt(final String classname, String methodName) {
		// Create a File object on the root of the directory
		// containing the class file
		File file = new File(classOutputFolder);
		
		try {
			// Convert File to a URL
			URL url = file.toURL(); // file:/classes/demo
			URL[] urls = new URL[] { url };

			// Create a new class loader with the directory
			ClassLoader loader = new URLClassLoader(urls);

			// Load in the class; Class.childclass should be located in
			// the directory file:/class/demo/
			Class thisClass = loader.loadClass(classname);

			Class params[] = {};
			Object paramsObj[] = {};
			Object instance = thisClass.newInstance();
			Method thisMethod = thisClass.getDeclaredMethod(methodName, params);

			// run the testAdd() method on the instance:
			thisMethod.invoke(instance, paramsObj);
		} catch (MalformedURLException e) {
		} catch (ClassNotFoundException e) {
		} catch (Exception ex) {
			ex.printStackTrace();
		}
	}

	// Deletes all files and subdirectories under dir.
	// Returns true if all deletions were successful.
	// If a deletion fails, the method stops attempting to delete and returns false.
	public static boolean deleteDir(File dir) {
	    if (dir.isDirectory()) {
	        String[] children = dir.list();
	        for (int i=0; i<children.length; i++) {
	            boolean success = deleteDir(new File(dir, children[i]));
	            if (!success) {
	                return false;
	            }
	        }
	    }

	    // The directory is now empty so delete it
	    return dir.delete();
	}
}